/* mohawk - small HTTP server
**
** Copyright © 2010 by Baptiste Daroussin <bapt@FreeBSD.org>,
** Copyright © 2010 by Freddy Dissaux <freddy.dsx@free.fr>,
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice, this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice, this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
** OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
*/
%{
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <err.h>
#define SYSLOG_NAMES 1
#include <syslog.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <netdb.h>
#include <ifaddrs.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <pwd.h>
#include "mohawk.h"

#define WANT_FILENAME 0
#define WANT_DIRNAME  1

#define ADDR_INFO_RESET             1
#define AUTH_PATH_RESET             2
#define AUTH_PATTERNS_RESET         4
#define CGI_ENV_RESET               8
#define CGI_MAP_RESET              16
#define CGI_PATTERNS_RESET         32
#define CHARSET_RESET              64
#define DIRLIST_CSS_URL_RESET     128
#define EMAIL_ADMIN_RESET         256
#define INDEX_NAMES_RESET         512
#define MOHAWK_NAME_RESET        1024
#define NO_AUTH_PATTERNS_RESET   2048
#define NO_CGI_MAPS_RESET        4096
#define NO_LOGS_PATTERNS_RESET   8192
#define RESTRICT_FROMS_RESET    16384
#define STATUS_URL_RESET        32768


#define CLONE_HASH(FIELD, MASK) \
	if (force || (reset_mask & MASK) != MASK) { \
		TAILQ_FOREACH(from_hash, &from->FIELD, entry) { \
			if ((to_hash = malloc(sizeof(struct mohawk_hash))) == NULL) \
				errx(1, "Unable to allocate memory"); \
			to_hash->key = from_hash->key; \
			to_hash->value = from_hash->value; \
			TAILQ_INSERT_TAIL(&to->FIELD, to_hash, entry); \
			if (MASK == CGI_ENV_RESET) \
				to->cgi_env_count++; \
		} \
	} 

#define CLONE_LIST(FIELD, MASK) \
	if (force || (reset_mask & MASK) != MASK) { \
		TAILQ_FOREACH(from_list, &from->FIELD, entry) { \
			if ((to_list = malloc(sizeof(struct mohawk_list))) == NULL) \
				errx(1, "Unable to allocate memory"); \
			to_list->item = from_list->item; \
			TAILQ_INSERT_TAIL(&to->FIELD, to_list, entry); \
		} \
	} 

#define NUMBER_IS_PORT(PORT) \
	if (PORT < 1 || PORT > IPPORT_MAX) { \
		yyerror("'%d': not a valid port", PORT); \
		YYERROR; \
	}

#define SET_MOHAWK_LIST(ITEM, FIELD) \
	list_mohawk->item = ITEM; \
	TAILQ_INSERT_TAIL(&current_vh->FIELD, list_mohawk, entry); \
	if ((list_mohawk = malloc(sizeof(struct mohawk_list))) == NULL) \
		errx(1, "Unable to allocate memory");

#define SET_MOHAWK_HASH(KEY, VALUE, FIELD) \
	hash_mohawk->key = KEY; \
	hash_mohawk->value = VALUE; \
	TAILQ_INSERT_TAIL(&current_vh->FIELD, hash_mohawk, entry); \
	if ((hash_mohawk = malloc(sizeof(struct mohawk_hash))) == NULL) \
		errx(1, "Unable to allocate memory");

#define RESET_MOHAWK_HASH(FIELD) \
	hash_mohawk = TAILQ_FIRST(&current_vh->FIELD); \
	while (hash_mohawk != NULL) { \
		hash_next = TAILQ_NEXT(hash_mohawk, entry); \
		free(hash_mohawk); \
		hash_mohawk = hash_next; \
	} \
	TAILQ_INIT(&current_vh->FIELD); \
	if ((hash_mohawk = malloc(sizeof(struct mohawk_hash))) == NULL) \
		errx(1, "Unable to allocate memory");

#define RESET_MOHAWK_LIST(FIELD) \
	list_mohawk = TAILQ_FIRST(&current_vh->FIELD); \
	while (list_mohawk != NULL) { \
		list_next = TAILQ_NEXT(list_mohawk, entry); \
		free(list_mohawk); \
		list_mohawk = list_next; \
	} \
	TAILQ_INIT(&current_vh->FIELD); \
	if ((list_mohawk = malloc(sizeof(struct mohawk_list))) == NULL) \
		errx(1, "Unable to allocate memory");

int yyparse(void);
void yyerror(const char *, ...);
int yywrap(void);
int word_is_fs(const char *, int);
int interface(const char *, const char *, int);
struct vhost* vhost_lookup(const char *);
struct vhost* vhost_new(void);
void vhost_clone(struct vhost *, struct vhost *, int);
void vhost_add(const char *, const char *);

extern int yylineno;

struct vhost *current_vh = NULL;
struct mohawk_hash *hash_mohawk = NULL, *hash_next = NULL;
struct mohawk_list *list_mohawk = NULL, *list_next = NULL;
int reset_mask = 0;


void
yyerror(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vprintf(fmt, ap);
	printf(" at line %d\n", yylineno);
	va_end(ap);
	exit(EXIT_FAILURE);
}

int yywrap(void)
{
	return 1;
} 

int
word_is_fs(const char *word, int want)
{
	struct stat sb;
	char buf[PATH_MAX];
	const char *path = word;

	if (conf.chroot != NULL) {
		snprintf(buf, sizeof(buf), "%s%s", conf.chroot, word); 
		path = buf;
	}

	if (stat(path, &sb) != 0) {
		yyerror("'%s': not a %s", path, want == WANT_DIRNAME ? "directory" : "file");
		return 1;
	}
	if (S_ISDIR(sb.st_mode) && want == WANT_DIRNAME)
		return 0;
	if (S_ISREG(sb.st_mode) && want == WANT_FILENAME)
		return 0;
	
	yyerror("'%s': not a %s", path, want == WANT_DIRNAME ? "directory" : "file");
	return 1;
}

const char *
facility_lookup(const char *name, int *facility)
{
	CODE* item = facilitynames;
	while (item->c_name && strcmp(item->c_name, name))
		item++;

	if (facility)
		*facility = item->c_name ? item->c_val >> 3 : -1;

	return item->c_name;
}

int
interface(const char *s, const char *service, int family)
{
	struct address_info *addr_info;
	struct ifaddrs *ifap, *p;
	struct addrinfo *ai, *current_ai;
	struct addrinfo hints;
	const struct sockaddr_in6 *sin6 = NULL;
	char hostname[MAXHOSTNAMELEN];
	int error, ret = 0, sa_family = 0;
	void *fix_warnings;

	if (getifaddrs(&ifap) == -1)
		errx(EXIT_FAILURE, "getifaddrs() return -1");

	for (p = ifap; p != NULL; p = p->ifa_next) {
		if (p->ifa_addr == NULL)
			continue;
		if (!(p->ifa_flags & IFF_UP))
			continue;
		if (s != NULL && strcmp(s, p->ifa_name) != 0)
			continue;

		sa_family = p->ifa_addr->sa_family;
		if (sa_family != AF_INET && sa_family != AF_INET6)
			continue;
		if (family > 0 && sa_family != family)
			continue;

		// yes, we skip local link address
		if (sa_family == AF_INET6) {
			fix_warnings = p->ifa_addr;
			sin6 = (const struct sockaddr_in6 *)fix_warnings;
			if (IN6_IS_ADDR_LINKLOCAL(&sin6->sin6_addr))
				continue;
		}

		memset(&hints, 0, sizeof hints);
		hints.ai_family = PF_UNSPEC;
		hints.ai_flags = AI_PASSIVE;
		hints.ai_socktype = SOCK_STREAM;

		if ((error = getnameinfo(p->ifa_addr, p->ifa_addr->sa_len,
		    hostname, MAXHOSTNAMELEN, NULL, 0, NI_NUMERICHOST)) != 0) {
#ifdef USE_DEBUG
			warnx("warning only, getnameinfo(): %s", gai_strerror(error));
#endif
			ret = 0;
			break;
		}

		if ((error = getaddrinfo(hostname, service, &hints, &ai)) != 0) {
#ifdef USE_DEBUG
			warnx("warning only, interface - getaddrinfo('%s', '%s'): %s", hostname, service, gai_strerror(error));
#endif
			ret = 0;
			break;
		}

		for (current_ai = ai; current_ai != NULL; current_ai = current_ai->ai_next) {
			if ((addr_info = calloc(1, sizeof(struct address_info))) == NULL)
				errx(EXIT_FAILURE, "Unable to allocate memory");
			addr_info->ai = current_ai;
			SLIST_INSERT_HEAD(&current_vh->address_infos, addr_info, entry);
			ret++;
		}
	}
	freeifaddrs(ifap);
	return ret;
}

int
host(const char *hostname, const char *service)
{
	struct address_info *addr_info;
	struct addrinfo *ai, *current_ai;
	struct addrinfo hints;
	int error, ret = 0;

	if (strcmp(hostname, "0.0.0.0") == 0)
		return interface(NULL, service, AF_INET);
	if (strcmp(hostname, "::") == 0)
		return interface(NULL, service, AF_INET6);

	memset(&hints, 0, sizeof hints);
	hints.ai_family = PF_UNSPEC;
	hints.ai_flags = AI_PASSIVE;
	hints.ai_socktype = SOCK_STREAM;

	if ((error = getaddrinfo(hostname, service, &hints, &ai)) != 0) {
#ifdef USE_DEBUG
		warnx("warning only, host - getaddrinfo('%s', '%s'): %s", hostname, service, gai_strerror(error));
#endif
		return 0;
	}

	for (current_ai = ai; current_ai != NULL; current_ai = current_ai->ai_next) {
		if ((addr_info = calloc(1, sizeof(struct address_info))) == NULL)
			errx(EXIT_FAILURE, "Unable to allocate memory");
		addr_info->ai = current_ai;
		SLIST_INSERT_HEAD(&current_vh->address_infos, addr_info, entry);
		ret++;
	}
	return ret;
}

struct vhost *
vhost_new(void)
{
	struct vhost *vh;
	char buf[64]; //magic number. sorry :(

	if ((vh = malloc(sizeof(struct vhost))) == NULL)
		errx(EXIT_FAILURE, "Unable to allocate memory");

	TAILQ_INIT(&vh->auth_patterns);
	TAILQ_INIT(&vh->cgi_patterns);
	TAILQ_INIT(&vh->index_names);
	TAILQ_INIT(&vh->no_auth_patterns);
	TAILQ_INIT(&vh->no_cgi_maps);
	TAILQ_INIT(&vh->no_log_patterns);
	TAILQ_INIT(&vh->restrict_froms);
	TAILQ_INIT(&vh->cgi_env);
	TAILQ_INIT(&vh->cgi_map);
	SLIST_INIT(&vh->address_infos);

	vh->auth_path = NULL;
	vh->charset = NULL;
	vh->dirlist_css_url = NULL;
	vh->email_admin = NULL;
	vh->host = NULL;
	snprintf(buf, sizeof(buf), "%s/%s", getprogname(), MOHAWK_VERSION);
	vh->mohawk_name = strdup(buf);
	vh->status_url = NULL;

	vh->authentication = -1;
	vh->cgi_expose_mohawk_version = -1;
	vh->cgi_env_count = 0;
	vh->dirlist = -1;
	vh->maxage = -1;
	vh->x_forwarded_for = -1;

	return vh;
}

struct vhost *
vhost_lookup(const char *hostname)
{
	struct vhost *vh;
	SLIST_FOREACH(vh, &conf.vhosts, entry) {
		if (strcmp(vh->host, hostname) == 0)
			return vh;
	}
	return NULL;
}

#define RESET_ADDR_INFO() \
	struct address_info *ai; \
	if (conf.default_vhost == NULL) { \
		yyerror("reset listen is forbidden inside default vhost"); \
		YYERROR; \
	} \
	while (!SLIST_EMPTY(&current_vh->address_infos)) { \
		ai = SLIST_FIRST(&current_vh->address_infos); \
		SLIST_REMOVE_HEAD(&current_vh->address_infos, entry); \
		free(ai); \
	}

#define SET_OR_MERGE_CHAR(FIELD, MASK) \
	if (!force && from->FIELD && !to->FIELD) \
		to->FIELD = from->FIELD; \
	if (force && from->FIELD) \
		to->FIELD = from->FIELD;

#define SET_OR_MERGE_INT(FIELD) \
	if (!force && from->FIELD && to->FIELD < 0) \
		to->FIELD = from->FIELD; \
	if (force && from->FIELD >= 0) \
		to->FIELD = from->FIELD;

void
vhost_clone(struct vhost *from, struct vhost *to, int force)
{
	struct mohawk_list *from_list, *to_list;
	struct mohawk_hash *from_hash, *to_hash;
	struct address_info *from_ai, *to_ai;

	// rootdir can't be reset. it's inherite or define
	if ((force || !to->rootdir) && from->rootdir)
		to->rootdir = from->rootdir;

	SET_OR_MERGE_CHAR(auth_path, AUTH_PATH_RESET);
	SET_OR_MERGE_CHAR(charset, CHARSET_RESET);
	SET_OR_MERGE_CHAR(dirlist_css_url, DIRLIST_CSS_URL_RESET);
	SET_OR_MERGE_CHAR(email_admin, EMAIL_ADMIN_RESET);
	SET_OR_MERGE_CHAR(mohawk_name, MOHAWK_NAME_RESET);
	SET_OR_MERGE_CHAR(status_url, CHARSET_RESET);

	SET_OR_MERGE_INT(authentication);
	SET_OR_MERGE_INT(cgi_expose_mohawk_version);
	SET_OR_MERGE_INT(dirlist);
	SET_OR_MERGE_INT(maxage);
	SET_OR_MERGE_INT(x_forwarded_for);

	CLONE_LIST(auth_patterns, AUTH_PATTERNS_RESET);
	CLONE_LIST(cgi_patterns, CGI_PATTERNS_RESET);
	CLONE_LIST(index_names, INDEX_NAMES_RESET);
	CLONE_LIST(no_auth_patterns, NO_AUTH_PATTERNS_RESET);
	CLONE_LIST(no_cgi_maps, NO_CGI_MAPS_RESET);
	CLONE_LIST(no_log_patterns, NO_LOGS_PATTERNS_RESET);
	CLONE_LIST(restrict_froms, RESTRICT_FROMS_RESET);

	CLONE_HASH(cgi_map, CGI_MAP_RESET);
	CLONE_HASH(cgi_env, CGI_ENV_RESET);

	if (force || (reset_mask & ADDR_INFO_RESET) != ADDR_INFO_RESET) {
		SLIST_FOREACH(from_ai, &from->address_infos, entry) {
			if ((to_ai = malloc(sizeof(struct address_info))) == NULL)
				errx(EXIT_FAILURE, "Unable to allocate memory");
			to_ai->ai = from_ai->ai;
			SLIST_INSERT_HEAD(&to->address_infos, to_ai, entry);
		}
	}

	return;
}

void
vhost_add(const char *s, const char *clone)
{
	char buf[PATH_MAX];
	struct vhost *vh_clone = NULL, *vh = NULL;
	int is_default = 0, nb = 0;

	if (strcmp("default", s) == 0) {
		is_default = 1;
		if (SLIST_EMPTY(&current_vh->address_infos)) {
			nb += interface(NULL, DEFAULT_HTTP_PORT, AF_INET);
			nb += interface(NULL, DEFAULT_HTTP_PORT, AF_INET6);
			if (!nb)
				errx(EXIT_FAILURE, "Can't find any ip for listenning on '%s'. Hu ?", DEFAULT_HTTP_PORT);
		}
	}

	if (!is_default && conf.default_vhost == NULL)
		errx(EXIT_FAILURE, "Call add_vhost() but no 'default' vhost found");

	vh_clone = conf.default_vhost;
	if (clone != NULL) {
		if ((vh_clone = vhost_lookup(clone)) == NULL)
			errx(EXIT_FAILURE, "Can't clone '%s': not found", clone);
	}

	vh = vhost_new();
	if (!is_default)
		vhost_clone(vh_clone, vh, 0);
	vhost_clone(current_vh, vh, 1);

	if (vh->rootdir == NULL)
		errx(EXIT_FAILURE, "Can't add '%s': no rootdir", s);

	if (SLIST_EMPTY(&vh->address_infos))
		errx(EXIT_FAILURE, "Can't add vhost '%s': no listen", s);

	if (vh->authentication > 0) {
		if (vh->auth_path == NULL)
			errx(EXIT_FAILURE, "Can't add vhost '%s': authentication is 'on' but no auth_path defined", s);
		
		if (strstr(vh->auth_path, vh->rootdir) != NULL)
			errx(EXIT_FAILURE, "Can't add vhost '%s': auth_path '%s' inside rootdir '%s'", s, vh->auth_path, vh->rootdir);

		if (snprintf(buf, PATH_MAX, "%s/%s", vh->auth_path, DEFAULT_AUTH_FILE) >= PATH_MAX)
			errx(EXIT_FAILURE, "Can't add vhost '%s': auth_path filename too long", s);

		if ((vh->auth_path = strdup(buf)) == NULL)
			errx(EXIT_FAILURE, "Can't strdup auth_path");
	}

	if (is_default) {
		vh->host = "*";
		conf.default_vhost = vh;
	}
	else {
		vh->host = s;
		SLIST_INSERT_HEAD(&conf.vhosts, vh, entry);
	}

	current_vh = vhost_new();
	reset_mask = 0;
	return;
}

void
parse_config(const char *filename)
{
	int fd;

	TAILQ_INIT(&conf.mime_type);
	if ((hash_mohawk = malloc(sizeof(struct mohawk_hash))) == NULL)
		errx(EXIT_FAILURE, "Unable to allocate memory");

	if ((list_mohawk = malloc(sizeof(struct mohawk_list))) == NULL)
		errx(EXIT_FAILURE, "Unable to allocate memory");

	conf.default_vhost = NULL;
	current_vh = vhost_new();

	if ((fd = open(filename, O_RDONLY)) == -1)
		err(EXIT_FAILURE, "%s", filename);

	if (close(STDIN_FILENO) == -1)
		err(EXIT_FAILURE, "stdin");

	if (dup(fd) == -1)
		err(EXIT_FAILURE, "%s", filename);

	yyparse();
	close(fd);

	if (conf.default_vhost == NULL)
		errx(EXIT_FAILURE, "No 'default' vhost found");
}

%}

%{
int yylex(void);
%}

/* syntax */
%token EBRACE OBRACE PORT 

/* mohawk */
%token CHROOT MIME_TYPE MOHAWK_DEBUG PIDFILE SYSLOG_FACILITY USER 

/* vhost */
%token VHOST CLONE

/* on/off */
%token AUTHENTICATION CGI_EXPOSE_MOHAWK_VERSION DIRLIST X_FORWARDED_FOR 

/* options */
%token AUTH_PATH CHARSET DIRLIST_CSS_URL EMAIL_ADMIN MAXAGE MOHAWK_NAME ROOTDIR STATUS_URL 

/* list */
%token AUTH_PATTERNS CGI_PATTERNS INDEX_NAMES NO_AUTH_PATTERNS NO_CGI_MAPS NO_LOG_PATTERNS RESTRICT_FROMS 

/* hash */
%token CGI_ENV CGI_MAP 

/* command */
%token LISTEN RESET 

/* values */
%token DEFAULT WORD NUMBER WORDS STATE

%union 
{
	int number;
	char *string;
}

%token <number> STATE
%token <number> NUMBER
%token <string> WORD
%token <string> WORDS

%%
commands: /* empty */
        | commands command
        ;

command: chroot | debug | mime_type | pidfile | syslog_facility | user | vhost ;

chroot: CHROOT WORD {
	if (word_is_fs($2, WANT_DIRNAME) != 0)
		YYERROR;
	conf.chroot = $2;
};

debug: MOHAWK_DEBUG STATE {
	if (conf.debug < 0)
		conf.debug = $2;
};

mime_type: MIME_TYPE mime_type_content { };

mime_type_content: OBRACE mime_type_items EBRACE { };

mime_type_items: | mime_type_items mime_type_item ;

mime_type_item: WORD WORD {
	hash_mohawk->key = $1;
	hash_mohawk->value = $2;
	TAILQ_INSERT_TAIL(&conf.mime_type, hash_mohawk, entry);
	if ((hash_mohawk = malloc(sizeof(struct mohawk_hash))) == NULL)
		errx(1, "Unable to allocate memory");
};

pidfile: PIDFILE WORD {
	if (conf.pidfile == NULL)
		conf.pidfile = $2;
};

syslog_facility: SYSLOG_FACILITY WORD {
	if ((conf.syslog_facility = facility_lookup($2, NULL)) == NULL) {
		yyerror("'%s': not a valid syslog facility name", $2);
		YYERROR;
	}
};

user: USER WORD {
	if (getpwnam($2) == NULL) {
		yyerror("'%s': not a valid user", $2);
		YYERROR;
	}
	conf.user = $2;
};

vhost: VHOST DEFAULT vhost_content { vhost_add("default", NULL); } 
     | VHOST WORD vhost_content { vhost_add($2, NULL); }
     | VHOST WORD CLONE WORD vhost_content { vhost_add($2, $4); };

vhost_content: OBRACE vhost_items EBRACE { };

vhost_items: | vhost_items vhost_item;

vhost_item: authentication | cgi_expose_mohawk_version | dirlist | x_forwarded_for
          | auth_path | charset | dirlist_css_url | email_admin | maxage | mohawk_name | rootdir | status_url
          | auth_patterns | cgi_patterns | index_names | no_auth_patterns | no_cgi_maps | no_log_patterns | restrict_froms
          | cgi_env | cgi_map
          | listen | reset

authentication: AUTHENTICATION STATE {
	current_vh->authentication = $2;
};

cgi_expose_mohawk_version: CGI_EXPOSE_MOHAWK_VERSION STATE {
	current_vh->cgi_expose_mohawk_version = $2;
};

dirlist: DIRLIST STATE {
	current_vh->dirlist = $2;
};

x_forwarded_for: X_FORWARDED_FOR STATE {
	current_vh->x_forwarded_for = $2;
};

auth_path: AUTH_PATH WORD {
	if (word_is_fs($2, WANT_DIRNAME) != 0)
		YYERROR;
	current_vh->auth_path = $2;
};

charset: CHARSET WORD {
	current_vh->charset = $2;
};

dirlist_css_url: DIRLIST_CSS_URL WORD {
	current_vh->dirlist_css_url = $2;
};

email_admin: EMAIL_ADMIN WORD {
	current_vh->email_admin = $2;
};

maxage: MAXAGE NUMBER {
	current_vh->maxage = $2;
};

mohawk_name: MOHAWK_NAME WORD { current_vh->mohawk_name = $2; }
           | MOHAWK_NAME WORDS { current_vh->mohawk_name = $2; };

rootdir: ROOTDIR WORD {
	if (word_is_fs($2, WANT_DIRNAME) != 0)
		YYERROR;
	current_vh->rootdir = $2;
};

status_url: STATUS_URL WORD {
	current_vh->status_url = $2;
};

auth_patterns: AUTH_PATTERNS auth_patterns_content { };

auth_patterns_content: OBRACE auth_patterns_items EBRACE { };

auth_patterns_items: | auth_patterns_items auth_patterns_item ;

auth_patterns_item: WORD { SET_MOHAWK_LIST($1, auth_patterns); };

cgi_patterns: CGI_PATTERNS cgi_patterns_content { };

cgi_patterns_content: OBRACE cgi_patterns_items EBRACE { };

cgi_patterns_items: | cgi_patterns_items cgi_patterns_item ;

cgi_patterns_item: WORD { SET_MOHAWK_LIST($1, cgi_patterns); };

index_names: INDEX_NAMES index_names_content { };

index_names_content: OBRACE index_names_items EBRACE { };

index_names_items: | index_names_items index_names_item ;

index_names_item: WORD { SET_MOHAWK_LIST($1, index_names); };

no_auth_patterns: NO_AUTH_PATTERNS no_auth_patterns_content { };

no_auth_patterns_content: OBRACE no_auth_patterns_items EBRACE { };

no_auth_patterns_items: | no_auth_patterns_items no_auth_patterns_item ;

no_auth_patterns_item: WORD { SET_MOHAWK_LIST($1, no_auth_patterns); };

no_cgi_maps: NO_CGI_MAPS no_cgi_maps_content { };

no_cgi_maps_content: OBRACE no_cgi_maps_items EBRACE { };

no_cgi_maps_items: | no_cgi_maps_items no_cgi_maps_item ;

no_cgi_maps_item: WORD { SET_MOHAWK_LIST($1, no_cgi_maps); };

no_log_patterns: NO_LOG_PATTERNS no_log_patterns_content { };

no_log_patterns_content: OBRACE no_log_patterns_items EBRACE { };

no_log_patterns_items: | no_log_patterns_items no_log_patterns_item ;

no_log_patterns_item: WORD { SET_MOHAWK_LIST($1, no_log_patterns); };

restrict_froms: RESTRICT_FROMS restrict_froms_content { };

restrict_froms_content: OBRACE restrict_froms_items EBRACE { };

restrict_froms_items: | restrict_froms_items restrict_froms_item ;

restrict_froms_item: WORD { SET_MOHAWK_LIST($1, restrict_froms); };

cgi_env: CGI_ENV cgi_env_content { };

cgi_env_content: OBRACE cgi_env_items EBRACE { };

cgi_env_items: | cgi_env_items cgi_env_item ;

cgi_env_item: WORD WORD { SET_MOHAWK_HASH($1, $2, cgi_env); current_vh->cgi_env_count++; }
            | WORD WORDS { SET_MOHAWK_HASH($1, $2, cgi_env); current_vh->cgi_env_count++; };

cgi_map: CGI_MAP cgi_map_content { };

cgi_map_content: OBRACE cgi_map_items EBRACE { };

cgi_map_items: | cgi_map_items cgi_map_item ;

cgi_map_item: WORD WORD { SET_MOHAWK_HASH($1, $2, cgi_map); }
            | WORD WORDS { SET_MOHAWK_HASH($1, $2, cgi_map); };

listen: LISTEN STATE WORD PORT NUMBER {
			NUMBER_IS_PORT($5);
			char buf[6];
			int nb = 0;
			snprintf(buf, sizeof(buf), "%d", $5);
			nb = interface($3, buf, -1) + host($3, buf);
			if (!nb) {
				yylineno--;
				yyerror("invalid virtual ip or interface or hostname: '%s' '%d'", $3, $5);
				YYERROR;
			}
        }
      | LISTEN STATE WORD PORT WORD {
			int nb = 0;
			nb = interface($3, $5, -1) + host($3, $5);
			if (!nb) {
				yylineno--;
				yyerror("invalid virtual ip or interface or hostname: '%s' '%s'", $3, $5);
				YYERROR;
			}
        }
      | LISTEN STATE PORT NUMBER {
			NUMBER_IS_PORT($4);
			char buf[6];
			int nb = 0;
			snprintf(buf, sizeof(buf), "%d", $4);
			nb = interface(NULL, buf, AF_INET) + interface(NULL, buf, AF_INET6);
			if (!nb) {
				yylineno--;
				yyerror("invalid port: '%d'", $4);
				YYERROR;
			}
        }
      | LISTEN STATE PORT WORD {
			int nb = 0;
			nb = interface(NULL, $4, AF_INET) + interface(NULL, $4, AF_INET6);
			if (!nb) {
				yylineno--;
				yyerror("invalid port: '%s'", $4);
				YYERROR;
			}
        }
      | LISTEN STATE WORD {
			int nb = 0;
			nb = interface($3, DEFAULT_HTTP_PORT, -1) + host($3, DEFAULT_HTTP_PORT);
			if (!nb) {
				yylineno--;
				yyerror("invalid virtual ip or interface or hostname: '%s'", $3);
				YYERROR;
			}
        };

reset: RESET AUTH_PATTERNS { RESET_MOHAWK_LIST(auth_patterns); reset_mask += AUTH_PATTERNS_RESET; }
     | RESET CGI_PATTERNS { RESET_MOHAWK_LIST(cgi_patterns); reset_mask += CGI_PATTERNS_RESET; }
     | RESET INDEX_NAMES { RESET_MOHAWK_LIST(index_names); reset_mask += INDEX_NAMES_RESET; }
     | RESET NO_CGI_MAPS { RESET_MOHAWK_LIST(no_cgi_maps); reset_mask += NO_CGI_MAPS_RESET; }
     | RESET NO_AUTH_PATTERNS { RESET_MOHAWK_LIST(no_auth_patterns); reset_mask += NO_AUTH_PATTERNS_RESET; }
     | RESET NO_LOG_PATTERNS { RESET_MOHAWK_LIST(no_log_patterns); reset_mask += NO_LOGS_PATTERNS_RESET; }
     | RESET RESTRICT_FROMS { RESET_MOHAWK_LIST(restrict_froms); reset_mask += RESTRICT_FROMS_RESET; }

     | RESET CGI_ENV { RESET_MOHAWK_HASH(cgi_env); current_vh->cgi_env_count = 0; reset_mask += CGI_ENV_RESET; }
     | RESET CGI_MAP { RESET_MOHAWK_HASH(cgi_map); reset_mask += CGI_MAP_RESET; }

     | RESET AUTH_PATH { current_vh->auth_path = NULL; reset_mask += AUTH_PATH_RESET; }
     | RESET CHARSET { current_vh->charset = NULL; reset_mask += CHARSET_RESET; }
     | RESET DIRLIST_CSS_URL { current_vh->dirlist_css_url = NULL; reset_mask += DIRLIST_CSS_URL_RESET; }
     | RESET EMAIL_ADMIN { current_vh->email_admin = NULL; reset_mask += EMAIL_ADMIN_RESET; }
     | RESET MOHAWK_NAME { current_vh->mohawk_name = NULL; reset_mask += MOHAWK_NAME_RESET; }
     | RESET STATUS_URL { current_vh->status_url = NULL; reset_mask += STATUS_URL_RESET; }

     | RESET CGI_EXPOSE_MOHAWK_VERSION { current_vh->cgi_expose_mohawk_version = -1; }
     | RESET MAXAGE { current_vh->maxage = -1; }
     | RESET X_FORWARDED_FOR { current_vh->x_forwarded_for = -1; }

     | RESET LISTEN { RESET_ADDR_INFO(); reset_mask += ADDR_INFO_RESET; };

%%

/* vim: set sw=4 sts=4 ts=4 : */
